"use strict";

exports.__esModule = true;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _redblack = require("redblack");

var RedBlackTree = _interopRequireWildcard(_redblack);

var _signals = require("signals");

var _CacheDependency = require("./CacheDependency");

var _CacheDependency2 = _interopRequireDefault(_CacheDependency);

var _Utils = require("./Utils");

var Utils = _interopRequireWildcard(_Utils);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Cache = function () {
    function Cache(maxSize) {
        _classCallCheck(this, Cache);

        this._map = Object.create(null);
        this._tree = RedBlackTree.tree();
        this._size = 0;
        this._lastOrder = 1;
        this.itemRemoved = new _signals.Signal();

        if (maxSize != null && maxSize <= 0) throw new Error("'maxSize' shoud be greater than 0");

        this.maxSize = maxSize;
    }

    _createClass(Cache, [{
        key: "set",
        value: function set(key, value, options) {
            var _this = this;

            options = options || {};

            if (value === void 0) throw new Error("'value' cannot be undefined");

            var deps = options.dependencies ? Array.isArray(options.dependencies) ? options.dependencies : [options.dependencies] : [];

            for (var i = 0; i < deps.length; i++) {
                if (!(deps[i] instanceof _CacheDependency2.default)) throw new Error("'dependencies' should be one or more 'CacheDependency' instances");
            }var absoluteExpiration = typeof options.absoluteExpiration === "function" ? options.absoluteExpiration.call(options) : options.absoluteExpiration;

            if (absoluteExpiration != null && !(absoluteExpiration instanceof Date)) throw new Error("'absoluteExpiration' is not a Date");

            if (options.slidingExpirationMsec != null) {
                if (absoluteExpiration != null) throw new Error("'slidingExpirationMsec' cannot be set if 'absoluteExpiration' is set");

                if (options.slidingExpirationMsec < 0 || options.slidingExpirationMsec > 31536000000) throw new Error("allowed range for 'slidingExpirationMsec' is [0..31536000000] (up to one year)");
            }

            var keyStruct = Utils.getKeyStruct(key);
            var prevItem = this._map[keyStruct.keyStr];

            if (prevItem) this._removeItem(prevItem, "Overwritten");

            var item = {
                order: this._lastOrder++,
                key: key,
                keyStr: keyStruct.keyStr,
                keyData: keyStruct.keyData,
                value: value,
                absoluteExpiration: absoluteExpiration,
                slidingExpirationMsec: options.slidingExpirationMsec,
                removeCallback: options.removeCallback
            };

            this._updateExpiration(item);

            if (this.maxSize != null && this.maxSize === this.getSize()) this._removeItem(this._getOldestItem(), "Underused");

            this._map[item.keyStr] = item;
            this._tree.insert(item.order, item);

            if (!prevItem) this._size++;

            var depTriggered = function depTriggered() {
                for (var _i = 0; _i < deps.length; _i++) {
                    deps[_i].triggered.remove(depTriggered);
                }_this.remove(key, "DependencyTriggered");
            };

            for (var _i2 = 0; _i2 < deps.length; _i2++) {
                var dependency = deps[_i2];

                dependency._attach(this);
                dependency.triggered.addOnce(depTriggered);
            }
        }
    }, {
        key: "get",
        value: function get(key) {
            var keyStruct = Utils.getKeyStruct(key);
            var item = this._map[keyStruct.keyStr];

            if (item !== void 0) {
                this._updateExpiration(item);
                return item.value;
            }
        }
    }, {
        key: "has",
        value: function has(key) {
            var keyStruct = Utils.getKeyStruct(key);
            return this._map[keyStruct.keyStr] !== void 0;
        }
    }, {
        key: "remove",
        value: function remove(key, reason) {
            var keys = Array.isArray(key) ? key : [key];

            for (var i = 0; i < keys.length; i++) {
                var keyStruct = Utils.getKeyStruct(keys[i]);
                var item = this._map[keyStruct.keyStr];

                this._removeItem(item, reason);
            }
        }
    }, {
        key: "clear",
        value: function clear(testFn) {
            var removedCount = 0;

            for (var keyStr in this._map) {
                var item = this._map[keyStr];

                if (!testFn || testFn(item.key, item.value) === true) {
                    this._removeItem(item);
                    removedCount++;
                }
            }

            return removedCount;
        }
    }, {
        key: "getSize",
        value: function getSize() {
            return this._size;
        }
    }, {
        key: "enumerate",
        value: function enumerate(fn) {
            for (var keyStr in this._map) {
                var item = this._map[keyStr];

                if (fn.call(this, item.key, item.value) === false) break;
            }
        }
    }, {
        key: "_getOldestItem",
        value: function _getOldestItem() {
            var node = this._tree.root;

            while (node.left !== null) {
                node = node.left;
            }return node.value;
        }
    }, {
        key: "_removeItem",
        value: function _removeItem(item, reason) {
            reason = reason || "Removed";

            if (typeof item.removeCallback === "function") item.removeCallback.apply(this, [item.key, item.value, reason]);

            delete this._map[item.keyStr];
            this._tree.delete(item.order);
            this._size--;

            this.itemRemoved.dispatch(item.key, item.value, reason);
        }
    }, {
        key: "_updateExpiration",
        value: function _updateExpiration(item) {
            var expires = void 0;

            if (item.expTimer != null) {
                clearTimeout(item.expTimer);
                item.expTimer = null;
            }

            if (item.absoluteExpiration != null) expires = item.absoluteExpiration;else if (item.slidingExpirationMsec != null) expires = new Date(new Date().getTime() + item.slidingExpirationMsec);

            if (expires) {
                var keyStr = item.keyStr;

                item.expTimer = setTimeout(this.remove(keyStr, "Expired"), expires.getTime() - new Date().getTime());
            }
        }
    }]);

    return Cache;
}();

exports.default = Cache;